/***************************************************************************
  qgswcscapabilities.cpp  -  WCS capabilities
                             -------------------
    begin                : 17 Mar, 2005
    copyright            : (C) 2005 by Brendan Morley
    email                : morb at ozemail dot com dot au

    wcs                  : 4/2012 Radim Blazek, based on qgswmsprovider.cpp

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "qgslogger.h"
#include "qgswcscapabilities.h"
#include "qgsowsconnection.h"

#include <cmath>

#include "qgsauthmanager.h"
#include "qgscoordinatetransform.h"
#include "qgsdatasourceuri.h"
#include "qgsrasterlayer.h"
#include "qgsrectangle.h"
#include "qgscoordinatereferencesystem.h"
#include "qgsnetworkaccessmanager.h"
#include "qgsmessageoutput.h"
#include "qgsmessagelog.h"
#include "qgscrscache.h"

#include <QNetworkRequest>
#include <QNetworkReply>
#include <QSet>
#include <QEventLoop>

#ifdef _MSC_VER
#include <float.h>
#define isfinite(x) _finite(x)
#endif

#ifdef QGISDEBUG
#include <QFile>
#include <QDir>
#endif

QgsWcsCapabilities::QgsWcsCapabilities( QgsDataSourceURI const &theUri )
    : mUri( theUri )
    , mCapabilitiesReply( nullptr )
    , mCoverageCount( 0 )
    , mCacheLoadControl( QNetworkRequest::PreferNetwork )
{
  QgsDebugMsg( "uri = " + mUri.encodedUri() );

  parseUri();

  retrieveServerCapabilities();
}

QgsWcsCapabilities::QgsWcsCapabilities()
    : mCapabilities()
    , mCapabilitiesReply( nullptr )
    , mCoverageCount( 0 )
    , mCacheLoadControl( QNetworkRequest::PreferNetwork )
{
}

QgsWcsCapabilities::~QgsWcsCapabilities()
{
  QgsDebugMsg( "deconstructing." );
}

void QgsWcsCapabilities::parseUri()
{
  mCacheLoadControl = QNetworkRequest::PreferNetwork;

  QString cache = mUri.param( "cache" );
  QgsDebugMsg( "cache = " + cache );

  if ( !cache.isEmpty() )
  {
    mCacheLoadControl = QgsNetworkAccessManager::cacheLoadControlFromName( cache );
  }
  QgsDebugMsg( QString( "mCacheLoadControl = %1" ).arg( mCacheLoadControl ) );
}

// TODO: return if successful
void QgsWcsCapabilities::setUri( QgsDataSourceURI const &theUri )
{
  mUri = theUri;

  clear();

  parseUri();

  retrieveServerCapabilities();
}

QString QgsWcsCapabilities::prepareUri( QString uri )
{
  if ( !uri.contains( '?' ) )
  {
    uri.append( '?' );
  }
  else if ( uri.right( 1 ) != "?" && uri.right( 1 ) != "&" )
  {
    uri.append( '&' );
  }

  return uri;
}

QgsWcsCapabilitiesProperty QgsWcsCapabilities::capabilities()
{
  return mCapabilities;
}

bool QgsWcsCapabilities::supportedCoverages( QVector<QgsWcsCoverageSummary> &coverageSummary )
{
  QgsDebugMsg( "Entering." );

  coverageSummary = mCoveragesSupported;

  QgsDebugMsg( "Exiting." );

  return true;
}

QString QgsWcsCapabilities::getCoverageUrl() const
{
  QString url = mCapabilities.getCoverageGetUrl;
  if ( url.isEmpty() )
  {
    url = mUri.param( "url" );
  }
  return url;
}

bool QgsWcsCapabilities::sendRequest( QString const & url )
{
  QgsDebugMsg( "url = " + url );
  mError = "";
  QNetworkRequest request( url );
  if ( !setAuthorization( request ) )
  {
    mError = tr( "Download of capabilities failed: network request update failed for authentication config" );
    QgsMessageLog::logMessage( mError, tr( "WCS" ) );
    return false;
  }
  request.setAttribute( QNetworkRequest::CacheSaveControlAttribute, true );
  request.setAttribute( QNetworkRequest::CacheLoadControlAttribute, mCacheLoadControl );
  QgsDebugMsg( QString( "mCacheLoadControl = %1" ).arg( mCacheLoadControl ) );

  QgsDebugMsg( QString( "getcapabilities: %1" ).arg( url ) );
  mCapabilitiesReply = QgsNetworkAccessManager::instance()->get( request );
  if ( !setAuthorizationReply( mCapabilitiesReply ) )
  {
    mCapabilitiesReply->deleteLater();
    mCapabilitiesReply = nullptr;
    mError = tr( "Download of capabilities failed: network reply update failed for authentication config" );
    QgsMessageLog::logMessage( mError, tr( "WCS" ) );
    return false;
  }

  connect( mCapabilitiesReply, SIGNAL( finished() ), this, SLOT( capabilitiesReplyFinished() ) );
  connect( mCapabilitiesReply, SIGNAL( downloadProgress( qint64, qint64 ) ), this, SLOT( capabilitiesReplyProgress( qint64, qint64 ) ) );

  QEventLoop loop;
  connect( this, SIGNAL( downloadFinished() ), &loop, SLOT( quit() ) );
  loop.exec( QEventLoop::ExcludeUserInputEvents );

  if ( mCapabilitiesResponse.isEmpty() )
  {
    if ( mError.isEmpty() )
    {
      mErrorFormat = "text/plain";
      mError = tr( "empty capabilities document" );
    }
    return false;
  }

  if ( mCapabilitiesResponse.startsWith( "<html>" ) ||
       mCapabilitiesResponse.startsWith( "<HTML>" ) )
  {
    mErrorFormat = "text/html";
    mError = mCapabilitiesResponse;
    return false;
  }
  return true;
}

void QgsWcsCapabilities::clear()
{
  mCoverageCount = 0;
  mCoveragesSupported.clear();
  mCapabilities = QgsWcsCapabilitiesProperty();
}

QString QgsWcsCapabilities::getCapabilitiesUrl( const QString& version ) const
{
  QString url = prepareUri( mUri.param( "url" ) ) + "SERVICE=WCS&REQUEST=GetCapabilities";

  if ( !version.isEmpty() )
  {
    // 1.0.0 - VERSION
    // 1.1.0 - AcceptVersions (not supported by UMN Mapserver 6.0.3 - defaults to latest 1.1
    if ( version.startsWith( "1.0" ) )
    {
      url += "&VERSION=" + version;
    }
    else if ( version.startsWith( "1.1" ) )
    {
      // Ignored by UMN Mapserver 6.0.3, see below
      url += "&AcceptVersions=" + version;
    }
  }
  return url;
}

QString QgsWcsCapabilities::getCapabilitiesUrl() const
{
  return getCapabilitiesUrl( mVersion );
}

bool QgsWcsCapabilities::retrieveServerCapabilities()
{
  clear();

  QStringList versions;

  QString preferredVersion = mUri.param( "version" );

  if ( !preferredVersion.isEmpty() )
  {
    versions << preferredVersion;
  }
  else
  {
    // We prefer 1.0 because 1.1 has many issues, each server implements it in defferent
    // way with various particularities
    // It may happen that server supports 1.1.0 but gives error for 1.1
    versions << "1.0.0" << "1.1.0,1.0.0";
  }

  Q_FOREACH ( const QString& v, versions )
  {
    if ( retrieveServerCapabilities( v ) )
    {
      return true;
    }
  }

  return false;
}

bool QgsWcsCapabilities::retrieveServerCapabilities( const QString& preferredVersion )
{
  clear();

  QString url = getCapabilitiesUrl( preferredVersion );

  if ( !sendRequest( url ) )
  {
    return false;
  }

  QgsDebugMsg( "Converting to Dom." );

  bool domOK;
  domOK = parseCapabilitiesDom( mCapabilitiesResponse, mCapabilities );

  if ( !domOK )
  {
    // We had an Dom exception -
    // mErrorTitle and mError are pre-filled by parseCapabilitiesDom

    mError += tr( "\nTried URL: %1" ).arg( url );

    QgsDebugMsg( "!domOK: " + mError );

    return false;
  }

  return true;
}

QString QgsWcsCapabilities::getDescribeCoverageUrl( QString const &identifier ) const
{
  QString url = prepareUri( mUri.param( "url" ) ) + "SERVICE=WCS&REQUEST=DescribeCoverage&VERSION=" + mVersion;

  if ( mVersion.startsWith( "1.0" ) )
  {
    url += "&COVERAGE=" + identifier;
  }
  else if ( mVersion.startsWith( "1.1" ) )
  {
    // in 1.1.0, 1.1.1, 1.1.2 the name of param is 'identifier'
    // but in KVP 'identifiers'
    url += "&IDENTIFIERS=" + identifier;
  }
  return url;
}

bool QgsWcsCapabilities::describeCoverage( QString const &identifier, bool forceRefresh )
{
  QgsDebugMsg( " identifier = " + identifier );

  QgsWcsCoverageSummary *coverage = coverageSummary( identifier );
  if ( !coverage )
  {
    QgsDebugMsg( "coverage not found" );
    return false;
  }

  if ( coverage->described && ! forceRefresh ) return true;

  QString url = getDescribeCoverageUrl( coverage->identifier );

  if ( !sendRequest( url ) )
  {
    return false;
  }

  QgsDebugMsg( "Converting to Dom." );

  bool domOK = false;
  if ( mVersion.startsWith( "1.0" ) )
  {
    domOK = parseDescribeCoverageDom10( mCapabilitiesResponse, coverage );
  }
  else if ( mVersion.startsWith( "1.1" ) )
  {
    domOK = parseDescribeCoverageDom11( mCapabilitiesResponse, coverage );
  }

  if ( !domOK )
  {
    // We had an Dom exception -
    // mErrorTitle and mError are pre-filled by parseCapabilitiesDom

    mError += tr( "\nTried URL: %1" ).arg( url );

    QgsDebugMsg( "!domOK: " + mError );

    return false;
  }

  return true;
}

void QgsWcsCapabilities::capabilitiesReplyFinished()
{
  if ( mCapabilitiesReply->error() == QNetworkReply::NoError )
  {
    QVariant redirect = mCapabilitiesReply->attribute( QNetworkRequest::RedirectionTargetAttribute );
    if ( !redirect.isNull() )
    {
      emit statusChanged( tr( "Capabilities request redirected." ) );

      QNetworkRequest request( redirect.toUrl() );
      if ( !setAuthorization( request ) )
      {
        mCapabilitiesResponse.clear();
        mError = tr( "Download of capabilities failed: network request update failed for authentication config" );
        QgsMessageLog::logMessage( mError, tr( "WCS" ) );
        return;
      }
      request.setAttribute( QNetworkRequest::CacheLoadControlAttribute, QNetworkRequest::PreferNetwork );
      request.setAttribute( QNetworkRequest::CacheSaveControlAttribute, true );

      mCapabilitiesReply->deleteLater();
      QgsDebugMsg( QString( "redirected getcapabilities: %1" ).arg( redirect.toString() ) );
      mCapabilitiesReply = QgsNetworkAccessManager::instance()->get( request );
      if ( !setAuthorizationReply( mCapabilitiesReply ) )
      {
        mCapabilitiesResponse.clear();
        mCapabilitiesReply->deleteLater();
        mCapabilitiesReply = nullptr;
        mError = tr( "Download of capabilities failed: network reply update failed for authentication config" );
        QgsMessageLog::logMessage( mError, tr( "WCS" ) );
        return;
      }

      connect( mCapabilitiesReply, SIGNAL( finished() ), this, SLOT( capabilitiesReplyFinished() ) );
      connect( mCapabilitiesReply, SIGNAL( downloadProgress( qint64, qint64 ) ), this, SLOT( capabilitiesReplyProgress( qint64, qint64 ) ) );
      return;
    }

    mCapabilitiesResponse = mCapabilitiesReply->readAll();

    if ( mCapabilitiesResponse.isEmpty() )
    {
      mErrorFormat = "text/plain";
      mError = tr( "empty of capabilities: %1" ).arg( mCapabilitiesReply->errorString() );
    }
  }
  else
  {
    // Resend request if AlwaysCache
    QNetworkRequest request = mCapabilitiesReply->request();
    if ( request.attribute( QNetworkRequest::CacheLoadControlAttribute ).toInt() == QNetworkRequest::AlwaysCache )
    {
      QgsDebugMsg( "Resend request with PreferCache" );
      request.setAttribute( QNetworkRequest::CacheLoadControlAttribute, QNetworkRequest::PreferCache );

      mCapabilitiesReply->deleteLater();

      mCapabilitiesReply = QgsNetworkAccessManager::instance()->get( request );
      if ( !setAuthorizationReply( mCapabilitiesReply ) )
      {
        mCapabilitiesResponse.clear();
        mCapabilitiesReply->deleteLater();
        mCapabilitiesReply = nullptr;
        mError = tr( "Download of capabilities failed: network reply update failed for authentication config" );
        QgsMessageLog::logMessage( mError, tr( "WCS" ) );
        return;
      }
      connect( mCapabilitiesReply, SIGNAL( finished() ), this, SLOT( capabilitiesReplyFinished() ) );
      connect( mCapabilitiesReply, SIGNAL( downloadProgress( qint64, qint64 ) ), this, SLOT( capabilitiesReplyProgress( qint64, qint64 ) ) );
      return;
    }

    mErrorFormat = "text/plain";
    mError = tr( "Download of capabilities failed: %1" ).arg( mCapabilitiesReply->errorString() );
    QgsMessageLog::logMessage( mError, tr( "WCS" ) );
    mCapabilitiesResponse.clear();
  }

  mCapabilitiesReply->deleteLater();
  mCapabilitiesReply = nullptr;

  emit downloadFinished();
}

void QgsWcsCapabilities::capabilitiesReplyProgress( qint64 bytesReceived, qint64 bytesTotal )
{
  QString msg = tr( "%1 of %2 bytes of capabilities downloaded." ).arg( bytesReceived ).arg( bytesTotal < 0 ? QString( "unknown number of" ) : QString::number( bytesTotal ) );
  QgsDebugMsg( msg );
  emit statusChanged( msg );
}

QString QgsWcsCapabilities::stripNS( const QString & name )
{
  return name.contains( ':' ) ? name.section( ':', 1 ) : name;
}

bool QgsWcsCapabilities::parseCapabilitiesDom( QByteArray const &xml, QgsWcsCapabilitiesProperty &capabilities )
{
  QgsDebugMsg( "Entered." );
#ifdef QGISDEBUG
  QFile file( QDir::tempPath() + "/qgis-wcs-capabilities.xml" );
  if ( file.open( QIODevice::WriteOnly ) )
  {
    file.write( xml );
    file.close();
  }
#endif

  if ( ! convertToDom( xml ) ) return false;

  QDomElement docElem = mCapabilitiesDom.documentElement();

  // Assert that the DTD is what we expected (i.e. a WCS Capabilities document)
  QgsDebugMsg( "testing tagName " + docElem.tagName() );

  QString tagName = stripNS( docElem.tagName() );
  if (
    // We don't support 1.0, but try WCS_Capabilities tag to get version
    tagName != "WCS_Capabilities" && // 1.0
    tagName != "Capabilities"  // 1.1, tags seen: Capabilities, wcs:Capabilities
  )
  {
    if ( tagName == "ExceptionReport" )
    {
      mErrorTitle = tr( "Exception" );
      mErrorFormat = "text/plain";
      mError = tr( "Could not get WCS capabilities: %1" ).arg( domElementText( docElem, "Exception.ExceptionText" ) );
    }
    else
    {
      mErrorTitle = tr( "Dom Exception" );
      mErrorFormat = "text/plain";
      mError = tr( "Could not get WCS capabilities in the expected format (DTD): no %1 found.\nThis might be due to an incorrect WCS Server URL.\nTag:%3\nResponse was:\n%4" )
               .arg( "Capabilities",
                     docElem.tagName(),
                     QString( xml ) );
    }

    QgsLogger::debug( "Dom Exception: " + mError );

    return false;
  }

  capabilities.version = docElem.attribute( "version" );
  mVersion = capabilities.version;

  if ( !mVersion.startsWith( "1.0" ) && !mVersion.startsWith( "1.1" ) )
  {
    mErrorTitle = tr( "Version not supported" );
    mErrorFormat = "text/plain";
    mError = tr( "WCS server version %1 is not supported by QGIS (supported versions: 1.0.0, 1.1.0, 1.1.2)" )
             .arg( mVersion );

    QgsLogger::debug( "WCS version: " + mError );

    return false;
  }

  if ( mVersion.startsWith( "1.0" ) )
  {
    capabilities.title = domElementText( docElem, "Service.name" );
    capabilities.abstract = domElementText( docElem, "Service.description" );
    // There is also "label" in 1.0

    capabilities.getCoverageGetUrl = domElement( docElem, "Capability.Request.GetCoverage.DCPType.HTTP.Get.OnlineResource" ).attribute( "xlink:href" );

    parseContentMetadata( domElement( docElem, "ContentMetadata" ), capabilities.contents );
  }
  else if ( mVersion.startsWith( "1.1" ) )
  {
    capabilities.title = domElementText( docElem, "ServiceIdentification.Title" );
    capabilities.abstract = domElementText( docElem, "ServiceIdentification.Abstract" );

    QList<QDomElement> operationElements = domElements( docElem, "OperationsMetadata.Operation" );
    Q_FOREACH ( const QDomElement& el, operationElements )
    {
      if ( el.attribute( "name" ) == "GetCoverage" )
      {
        capabilities.getCoverageGetUrl = domElement( el, "DCP.HTTP.Get" ).attribute( "xlink:href" );
      }
    }

    parseCoverageSummary( domElement( docElem, "Contents" ), capabilities.contents );
  }

  return true;
}

QDomElement QgsWcsCapabilities::firstChild( const QDomElement &element, const QString & name )
{
  QDomNode n1 = element.firstChild();
  while ( !n1.isNull() )
  {
    QDomElement el = n1.toElement();
    if ( !el.isNull() )
    {
      QString tagName = stripNS( el.tagName() );
      if ( tagName == name )
      {
        return el;
      }
    }
    n1 = n1.nextSibling();
  }
  return QDomElement();
}

QString QgsWcsCapabilities::firstChildText( const QDomElement &element, const QString & name )
{
  QDomElement el = firstChild( element, name );
  if ( !el.isNull() )
  {
    return el.text();
  }
  return QString();
}

QList<QDomElement> QgsWcsCapabilities::domElements( const QDomElement &element, const QString & path )
{
  QList<QDomElement> list;

  QStringList names = path.split( '.' );
  if ( names.isEmpty() ) return list;
  QString name = names.value( 0 );
  names.removeFirst();

  QDomNode n1 = element.firstChild();
  while ( !n1.isNull() )
  {
    QDomElement el = n1.toElement();
    if ( !el.isNull() )
    {
      QString tagName = stripNS( el.tagName() );
      if ( tagName == name )
      {
        if ( names.isEmpty() )
        {
          list.append( el );
        }
        else
        {
          list.append( domElements( el, names.join( "." ) ) );
        }
      }
    }
    n1 = n1.nextSibling();
  }

  return list;
}

QStringList QgsWcsCapabilities::domElementsTexts( const QDomElement &element, const QString &path )
{
  QStringList list;
  QList<QDomElement> elems = domElements( element, path );

  Q_FOREACH ( const QDomElement& el, elems )
  {
    list << el.text();
  }
  return list;
}

QDomElement QgsWcsCapabilities::domElement( const QDomElement &element, const QString & path )
{
  QStringList names = path.split( '.' );
  if ( names.isEmpty() ) return QDomElement();

  QDomElement el = firstChild( element, names.value( 0 ) );
  if ( names.size() == 1 || el.isNull() )
  {
    return el;
  }
  names.removeFirst();
  return domElement( el, names.join( "." ) );
}

QString QgsWcsCapabilities::domElementText( const QDomElement &element, const QString & path )
{
  QDomElement el = domElement( element, path );
  return el.text();
}

QList<int> QgsWcsCapabilities::parseInts( const QString &text )
{
  QList<int> list;
  Q_FOREACH ( const QString& s, text.split( ' ' ) )
  {
    bool ok;
    list.append( s.toInt( &ok ) );
    if ( !ok )
    {
      list.clear();
      return list;
    }
  }
  return list;
}

QList<double> QgsWcsCapabilities::parseDoubles( const QString &text )
{
  QList<double> list;
  Q_FOREACH ( const QString& s, text.split( ' ' ) )
  {
    bool ok;
    list.append( s.toDouble( &ok ) );
    if ( !ok )
    {
      list.clear();
      return list;
    }
  }
  return list;
}

QString QgsWcsCapabilities::crsUrnToAuthId( const QString &text )
{
  QString authid = text; // may be also non URN, for example 'EPSG:4326'

  // URN format: urn:ogc:def:objectType:authority:version:code
  // URN example: urn:ogc:def:crs:EPSG::4326
  QStringList urn = text.split( ':' );
  if ( urn.size() == 7 )
  {
    authid = urn.value( 4 ) + ':' + urn.value( 6 );
  }

  return authid;
}

// ------------------------ 1.0 ----------------------------------------------


void QgsWcsCapabilities::parseContentMetadata( QDomElement const & e, QgsWcsCoverageSummary &coverageSummary )
{
  QDomNode n1 = e.firstChild();
  while ( !n1.isNull() )
  {
    QDomElement el = n1.toElement();
    if ( !el.isNull() )
    {
      QString tagName = stripNS( el.tagName() );

      if ( tagName == "CoverageOfferingBrief" )
      {
        QgsWcsCoverageSummary subCoverageSummary;

        initCoverageSummary( subCoverageSummary );

        parseCoverageOfferingBrief( el, subCoverageSummary, &coverageSummary );

        subCoverageSummary.valid = true;

        coverageSummary.coverageSummary.push_back( subCoverageSummary );
      }
    }
    n1 = n1.nextSibling();
  }
}

void QgsWcsCapabilities::parseCoverageOfferingBrief( QDomElement const & e, QgsWcsCoverageSummary &coverageSummary, QgsWcsCoverageSummary *parent )
{
  Q_UNUSED( parent );
  coverageSummary.orderId = ++mCoverageCount;

  coverageSummary.identifier = firstChildText( e, "name" );
  coverageSummary.title = firstChildText( e, "label" );
  coverageSummary.abstract = firstChildText( e, "description" );

  QList<QDomElement> posElements = domElements( e, "lonLatEnvelope.pos" );
  if ( posElements.size() != 2 )
  {
    QgsDebugMsg( "Wrong number of pos elements" );
  }
  else
  {
    QList<double> low = parseDoubles( posElements.value( 0 ).text() );
    QList<double> high = parseDoubles( posElements.value( 1 ).text() );
    if ( low.size() == 2 && high.size() == 2 )
    {
      coverageSummary.wgs84BoundingBox = QgsRectangle( low[0], low[1], high[0], high[1] );
      QgsDebugMsg( "wgs84BoundingBox = " + coverageSummary.wgs84BoundingBox.toString() );
    }
  }

  if ( !coverageSummary.identifier.isEmpty() )
  {
    QgsDebugMsg( "add coverage " + coverageSummary.identifier + " to supported" );
    mCoveragesSupported.push_back( coverageSummary );
  }

  if ( !coverageSummary.coverageSummary.empty() )
  {
    mCoverageParentIdentifiers[ coverageSummary.orderId ] = QStringList() << coverageSummary.identifier << coverageSummary.title << coverageSummary.abstract;
  }
  QgsDebugMsg( QString( "coverage orderId = %1 identifier = %2" ).arg( coverageSummary.orderId ).arg( coverageSummary.identifier ) );
}

bool QgsWcsCapabilities::convertToDom( QByteArray const &xml )
{
  QgsDebugMsg( "Entered." );
  // Convert completed document into a Dom
  QString errorMsg;
  int errorLine;
  int errorColumn;
  bool contentSuccess = mCapabilitiesDom.setContent( xml, false, &errorMsg, &errorLine, &errorColumn );

  if ( !contentSuccess )
  {
    mErrorTitle = tr( "Dom Exception" );
    mErrorFormat = "text/plain";
    mError = tr( "Could not get WCS capabilities: %1 at line %2 column %3\nThis is probably due to an incorrect WCS Server URL.\nResponse was:\n\n%4" )
             .arg( errorMsg )
             .arg( errorLine )
             .arg( errorColumn )
             .arg( QString( xml ) );

    QgsLogger::debug( "Dom Exception: " + mError );

    return false;
  }
  return true;
}

bool QgsWcsCapabilities::parseDescribeCoverageDom10( QByteArray const &xml, QgsWcsCoverageSummary *coverage )
{
  QgsDebugMsg( "coverage->identifier = " + coverage->identifier );
  if ( ! convertToDom( xml ) ) return false;

  QDomElement docElem = mCapabilitiesDom.documentElement();

  QgsDebugMsg( "testing tagName " + docElem.tagName() );

  QString tagName = stripNS( docElem.tagName() );
  if ( tagName != "CoverageDescription" )
  {
    mErrorTitle = tr( "Dom Exception" );
    mErrorFormat = "text/plain";
    mError = tr( "Could not get WCS capabilities in the expected format (DTD): no %1 found.\nThis might be due to an incorrect WCS Server URL.\nTag:%3\nResponse was:\n%4" )
             .arg( "CoverageDescription",
                   docElem.tagName(),
                   QString( xml ) );

    QgsLogger::debug( "Dom Exception: " + mError );

    return false;
  }

  QDomElement coverageOfferingElement = firstChild( docElem, "CoverageOffering" );

  if ( coverageOfferingElement.isNull() ) return false;
  QDomElement supportedCRSsElement = firstChild( coverageOfferingElement, "supportedCRSs" );

  // requestResponseCRSs and requestCRSs + responseCRSs are alternatives
  coverage->supportedCrs = domElementsTexts( coverageOfferingElement, "supportedCRSs.requestResponseCRSs" );
  // TODO: requestCRSs, responseCRSs - must be then implemented also in provider
  //QgsDebugMsg( "supportedCrs = " + coverage->supportedCrs.join( "," ) );

  coverage->nativeCrs = domElementText( coverageOfferingElement, "supportedCRSs.nativeCRSs" );

  // may be GTiff, GeoTIFF, TIFF, GIF, ....
  coverage->supportedFormat = domElementsTexts( coverageOfferingElement, "supportedFormats.formats" );
  QgsDebugMsg( "supportedFormat = " + coverage->supportedFormat.join( "," ) );

  // spatialDomain and Grid/RectifiedGrid are optional according to specificationi.
  // If missing, we cannot get native resolution and size.
  QDomElement gridElement = domElement( coverageOfferingElement, "domainSet.spatialDomain.RectifiedGrid" );

  if ( gridElement.isNull() )
  {
    // Grid has also GridEnvelope from which we can get coverage size but it does not
    gridElement = domElement( coverageOfferingElement, "domainSet.spatialDomain.Grid" );
  }

  // If supportedCRSs.nativeCRSs is not defined we try to get it from RectifiedGrid
  if ( coverage->nativeCrs.isEmpty() )
  {
    coverage->nativeCrs = gridElement.attribute( "srsName" );
  }

  if ( !gridElement.isNull() )
  {
    QList<int> low = parseInts( domElementText( gridElement, "limits.GridEnvelope.low" ) );
    QList<int> high = parseInts( domElementText( gridElement, "limits.GridEnvelope.high" ) );
    if ( low.size() == 2 && high.size() == 2 )
    {
      // low/high are indexes in grid -> size is hight - low + 1
      double width = high[0] - low[0] + 1;
      double height = high[1] - low[1] + 1;
      if ( width > 0 && height > 0 )
      {
        coverage->width = width;
        coverage->height = height;
        coverage->hasSize = true;
      }
    }
    // RectifiedGrid has also gml:origin which we don't need I think (attention however
    // it should contain gml:Point but mapserver 6.0.3 / WCS 1.0.0 is using gml:pos instead)
    // RectifiedGrid also contains 2 gml:offsetVector which could be used to get resolution
    // but it should be sufficient to calc resolution from size

    // TODO: check if coverage is rotated, in that case probably treat as without size
    //       or recalc resolution from rotated grid to base CRS
  }

  QList<QDomElement> envelopeElements = domElements( coverageOfferingElement, "domainSet.spatialDomain.Envelope" );

  QgsDebugMsg( QString( "%1 envelopeElements found" ).arg( envelopeElements.size() ) );

  Q_FOREACH ( const QDomElement& el, envelopeElements )
  {
    QString srsName = el.attribute( "srsName" );

    QList<QDomElement> posElements = domElements( el, "pos" );
    if ( posElements.size() != 2 )
    {
      QgsDebugMsg( "Wrong number of pos elements" );
      continue;
    }

    QList<double> low = parseDoubles( posElements.value( 0 ).text() );
    QList<double> high = parseDoubles( posElements.value( 1 ).text() );
    if ( low.size() == 2 && high.size() == 2 )
    {
      QgsRectangle box( low[0], low[1], high[0], high[1] );
      coverage->boundingBoxes.insert( srsName, box );
      QgsDebugMsg( "Envelope: " + srsName + " : " + box.toString() );
    }
  }

  coverage->times = domElementsTexts( coverageOfferingElement, "domainSet.temporalDomain.timePosition" );

  QList<QDomElement> timePeriodElements = domElements( coverageOfferingElement, "domainSet.temporalDomain.timePeriod" );

  QgsDebugMsg( QString( "%1 timePeriod found" ).arg( timePeriodElements.size() ) );

  Q_FOREACH ( const QDomElement& el, timePeriodElements )
  {
    QString beginPosition = domElementText( el, "beginPosition" );
    QString endPosition = domElementText( el, "endPosition" );
    QString timeResolution = domElementText( el, "timeResolution" );
    // Format used in request
    QString time = beginPosition + '/' + endPosition;
    if ( !timeResolution.isEmpty() )
    {
      time += '/' + timeResolution;
    }
    coverage->times << time;
  }

  // Find native bounding box
  if ( !coverage->nativeCrs.isEmpty() )
  {
    Q_FOREACH ( const QString& srsName, coverage->boundingBoxes.keys() )
    {
      if ( srsName == coverage->nativeCrs )
      {
        coverage->nativeBoundingBox = coverage->boundingBoxes.value( srsName );
      }
    }
  }

  // NULL / no data values
  // TODO: handle multiple range sets
  Q_FOREACH ( const QString& text, domElementsTexts( coverageOfferingElement, "rangeSet.RangeSet.nullValue.singleValue" ) )
  {
    bool ok;
    double val = text.toDouble( &ok );
    if ( ok )
    {
      coverage->nullValues.append( val );
    }
  }

  coverage->described = true;

  return true;
}
// ------------------------ 1.1 ----------------------------------------------
bool QgsWcsCapabilities::parseDescribeCoverageDom11( QByteArray const &xml, QgsWcsCoverageSummary *coverage )
{
  QgsDebugMsg( "coverage->identifier = " + coverage->identifier );
  if ( ! convertToDom( xml ) ) return false;

  QDomElement docElem = mCapabilitiesDom.documentElement();

  QgsDebugMsg( "testing tagName " + docElem.tagName() );

  QString tagName = stripNS( docElem.tagName() );
  if ( tagName != "CoverageDescriptions" )
  {
    mErrorTitle = tr( "Dom Exception" );
    mErrorFormat = "text/plain";
    mError = tr( "Could not get WCS capabilities in the expected format (DTD): no %1 found.\nThis might be due to an incorrect WCS Server URL.\nTag:%3\nResponse was:\n%4" )
             .arg( "CoverageDescriptions",
                   docElem.tagName(),
                   QString( xml ) );

    QgsLogger::debug( "Dom Exception: " + mError );

    return false;
  }

  // Get image size, we can get it from BoundingBox with crs=urn:ogc:def:crs:OGC::imageCRS
  // but while at least one BoundingBox is mandatory, it does not have to be urn:ogc:def:crs:OGC::imageCRS
  // TODO: if BoundingBox with crs=urn:ogc:def:crs:OGC::imageCRS is not found,
  // we could calculate image size from GridCRS.GridOffsets (if available)
  QList<QDomElement> boundingBoxElements = domElements( docElem, "CoverageDescription.Domain.SpatialDomain.BoundingBox" );

  QgsDebugMsg( QString( "%1 BoundingBox found" ).arg( boundingBoxElements.size() ) );

  Q_FOREACH ( const QDomElement& el, boundingBoxElements )
  {
    QString authid = crsUrnToAuthId( el.attribute( "crs" ) );
    QList<double> low = parseDoubles( domElementText( el, "LowerCorner" ) );
    QList<double> high = parseDoubles( domElementText( el, "UpperCorner" ) );

    if ( low.size() != 2 && high.size() != 2 ) continue;

    if ( el.attribute( "crs" ) == "urn:ogc:def:crs:OGC::imageCRS" )
    {
      coverage->width = ( int )( high[0] - low[0] + 1 );
      coverage->height = ( int )( high[1] - low[1] + 1 );
      coverage->hasSize = true;
    }
    else
    {
      QgsRectangle box;
      QgsCoordinateReferenceSystem crs = QgsCRSCache::instance()->crsByOgcWmsCrs( authid );
      if ( crs.isValid() && crs.axisInverted() )
      {
        box = QgsRectangle( low[1], low[0], high[1], high[0] );
      }
      else
      {
        box = QgsRectangle( low[0], low[1], high[0], high[1] );
      }
      coverage->boundingBoxes.insert( authid, box );
      QgsDebugMsg( "crs: " + crs.authid() + ' ' + crs.description() + QString( " axisInverted = %1" ).arg( crs.axisInverted() ) );
      QgsDebugMsg( "BoundingBox: " + authid + " : " + box.toString() );
    }
  }
  QgsDebugMsg( QString( "width = %1 height = %2" ).arg( coverage->width ).arg( coverage->height ) );

  // Each georectified coverage should have GridCRS
  QDomElement gridCRSElement = domElement( docElem, "CoverageDescription.Domain.SpatialDomain.GridCRS" );

  if ( !gridCRSElement.isNull() )
  {
    QString crsUrn = firstChildText( gridCRSElement, "GridBaseCRS" );
    coverage->nativeCrs = crsUrnToAuthId( crsUrn );
    QgsDebugMsg( "nativeCrs = " + coverage->nativeCrs );

    // TODO: consider getting coverage size from GridOffsets (resolution)
    // if urn:ogc:def:crs:OGC::imageCRS BoundingBox was not found
  }

  coverage->times = domElementsTexts( docElem, "CoverageDescription.Domain.TemporalDomain.timePosition" );

  QList<QDomElement> timePeriodElements = domElements( docElem, "CoverageDescription.Domain.TemporalDomain.timePeriod" );

  QgsDebugMsg( QString( "%1 timePeriod found" ).arg( timePeriodElements.size() ) );

  Q_FOREACH ( const QDomElement& el, timePeriodElements )
  {
    QString beginPosition = domElementText( el, "beginTime" );
    QString endPosition = domElementText( el, "endTime" );
    QString timeResolution = domElementText( el, "timeResolution" );
    // Format used in request
    QString time = beginPosition + '/' + endPosition;
    if ( !timeResolution.isEmpty() )
    {
      time += '/' + timeResolution;
    }
    coverage->times << time;
  }

  // NULL / no data values
  // TODO: handle multiple fields / ranges (?)
  Q_FOREACH ( const QString& text, domElementsTexts( docElem, "CoverageDescription.Range.Field.NullValue" ) )
  {
    bool ok;
    double val = text.toDouble( &ok );
    if ( ok )
    {
      coverage->nullValues.append( val );
    }
  }

  QStringList formats = domElementsTexts( docElem, "CoverageDescription.SupportedFormat" );
  // There could be formats from GetCapabilities
  if ( !formats.isEmpty() )
  {
    coverage->supportedFormat = formats;
  }


  QStringList crss = domElementsTexts( docElem, "CoverageDescription.SupportedCRS" );
  QSet<QString> authids; // Set, in case one CRS is in more formats (URN, non URN)
  Q_FOREACH ( const QString& crs, crss )
  {
    authids.insert( crsUrnToAuthId( crs ) );
  }
  if ( !authids.isEmpty() )
  {
    coverage->supportedCrs = authids.toList();
  }

  coverage->described = true;

  return true;
}


void QgsWcsCapabilities::parseCoverageSummary( QDomElement const & e, QgsWcsCoverageSummary &coverageSummary, QgsWcsCoverageSummary *parent )
{
  coverageSummary.orderId = ++mCoverageCount;

  coverageSummary.identifier = firstChildText( e, "Identifier" );
  coverageSummary.title = firstChildText( e, "Title" );
  coverageSummary.abstract = firstChildText( e, "Abstract" );

  QDomNode n1 = e.firstChild();
  while ( !n1.isNull() )
  {
    QDomElement el = n1.toElement();
    if ( !el.isNull() )
    {
      QString tagName = stripNS( el.tagName() );
      QgsDebugMsg( tagName + " : " + el.text() );

      if ( tagName == "SupportedFormat" )
      {
        // image/tiff, ...
        // Formats may be here (UMN Mapserver) or may not (GeoServer)
        coverageSummary.supportedFormat << el.text();
      }
      else if ( tagName == "SupportedCRS" )
      {
        // TODO: SupportedCRS may be URL referencing a document
        coverageSummary.supportedCrs << crsUrnToAuthId( el.text() );
      }
      else if ( tagName == "WGS84BoundingBox" )
      {
        QList<double> low = parseDoubles( domElementText( el, "LowerCorner" ) );
        QList<double> high = parseDoubles( domElementText( el, "UpperCorner" ) );

        if ( low.size() == 2 && high.size() == 2 )
        {
          coverageSummary.wgs84BoundingBox = QgsRectangle( low[0], low[1], high[0], high[1] );
        }
      }
    }
    n1 = n1.nextSibling();
  }
  //QgsDebugMsg( "supportedFormat = " + coverageSummary.supportedFormat.join( "," ) );

  // We collected params to be inherited, do children
  n1 = e.firstChild();
  while ( !n1.isNull() )
  {
    QDomElement el = n1.toElement();
    if ( !el.isNull() )
    {
      QString tagName = stripNS( el.tagName() );

      if ( tagName == "CoverageSummary" )
      {
        QgsDebugMsg( "      Nested coverage." );

        QgsWcsCoverageSummary subCoverageSummary;

        initCoverageSummary( subCoverageSummary );

        // Inherit
        subCoverageSummary.supportedCrs = coverageSummary.supportedCrs;
        subCoverageSummary.supportedFormat = coverageSummary.supportedFormat;

        parseCoverageSummary( el, subCoverageSummary, &coverageSummary );
        subCoverageSummary.valid = true;

        coverageSummary.coverageSummary.push_back( subCoverageSummary );
      }
    }
    n1 = n1.nextSibling();
  }

  if ( parent && parent->orderId > 1 ) // ignore Contents to put them on top level
  {
    QgsDebugMsg( QString( "coverage orderId = %1 identifier = %2 has parent %3" ).arg( coverageSummary.orderId ).arg( coverageSummary.identifier ).arg( parent->orderId ) );
    mCoverageParents[ coverageSummary.orderId ] = parent->orderId;
  }

  if ( !coverageSummary.identifier.isEmpty() )
  {
    QgsDebugMsg( "add coverage " + coverageSummary.identifier + " to supported" );
    mCoveragesSupported.push_back( coverageSummary );
  }

  if ( !coverageSummary.coverageSummary.empty() )
  {
    mCoverageParentIdentifiers[ coverageSummary.orderId ] = QStringList() << coverageSummary.identifier << coverageSummary.title << coverageSummary.abstract;
  }
  QgsDebugMsg( QString( "coverage orderId = %1 identifier = %2" ).arg( coverageSummary.orderId ).arg( coverageSummary.identifier ) );

}

void QgsWcsCapabilities::coverageParents( QMap<int, int> &parents, QMap<int, QStringList> &parentNames ) const
{
  parents = mCoverageParents;
  parentNames = mCoverageParentIdentifiers;
}

void QgsWcsCapabilities::initCoverageSummary( QgsWcsCoverageSummary &coverageSummary )
{
  coverageSummary.valid = false;
  coverageSummary.described = false;
  coverageSummary.width = 0;
  coverageSummary.height = 0;
  coverageSummary.hasSize = false;
}

QString QgsWcsCapabilities::lastErrorTitle()
{
  return mErrorTitle;
}

QString QgsWcsCapabilities::lastError()
{
  QgsDebugMsg( "returning '" + mError  + "'." );
  return mError;
}

QString QgsWcsCapabilities::lastErrorFormat()
{
  return mErrorFormat;
}

bool QgsWcsCapabilities::setAuthorization( QNetworkRequest &request ) const
{
  if ( mUri.hasParam( "authcfg" ) && !mUri.param( "authcfg" ).isEmpty() )
  {
    return QgsAuthManager::instance()->updateNetworkRequest( request, mUri.param( "authcfg" ) );
  }
  else if ( mUri.hasParam( "username" ) && mUri.hasParam( "password" ) )
  {
    QgsDebugMsg( "setAuthorization " + mUri.param( "username" ) );
    request.setRawHeader( "Authorization", "Basic " + QString( "%1:%2" ).arg( mUri.param( "username" ), mUri.param( "password" ) ).toAscii().toBase64() );
  }
  return true;
}

bool QgsWcsCapabilities::setAuthorizationReply( QNetworkReply *reply ) const
{
  if ( mUri.hasParam( "authcfg" ) && !mUri.param( "authcfg" ).isEmpty() )
  {
    return QgsAuthManager::instance()->updateNetworkReply( reply, mUri.param( "authcfg" ) );
  }
  return true;
}

void QgsWcsCapabilities::showMessageBox( const QString& title, const QString& text )
{
  QgsMessageOutput *message = QgsMessageOutput::createMessageOutput();
  message->setTitle( title );
  message->setMessage( text, QgsMessageOutput::MessageText );
  message->showMessage();
}

QgsWcsCoverageSummary QgsWcsCapabilities::coverage( QString const & theIdentifier )
{
  QgsWcsCoverageSummary * cp = coverageSummary( theIdentifier );
  if ( cp ) return *cp;

  QgsWcsCoverageSummary c;
  initCoverageSummary( c );
  return c;
}

QgsWcsCoverageSummary* QgsWcsCapabilities::coverageSummary( QString const & theIdentifier, QgsWcsCoverageSummary* parent )
{
  QgsDebugMsgLevel( "theIdentifier = " + theIdentifier, 5 );
  if ( !parent )
  {
    parent = &( mCapabilities.contents );
  }

  for ( QVector<QgsWcsCoverageSummary>::iterator c = parent->coverageSummary.begin(); c != parent->coverageSummary.end(); ++c )
  {
    if ( c->identifier == theIdentifier )
    {
      return c;
    }
    else
    {
      // search sub coverages
      QgsWcsCoverageSummary * sc = coverageSummary( theIdentifier, c );
      if ( sc )
      {
        return sc;
      }
    }
  }
  return nullptr;
}

QList<QgsWcsCoverageSummary> QgsWcsCapabilities::coverages()
{
  return coverageSummaries();
}

QList<QgsWcsCoverageSummary> QgsWcsCapabilities::coverageSummaries( QgsWcsCoverageSummary* parent )
{
  QList<QgsWcsCoverageSummary> list;
  if ( !parent )
  {
    parent = &( mCapabilities.contents );
  }

  for ( QVector<QgsWcsCoverageSummary>::iterator c = parent->coverageSummary.begin(); c != parent->coverageSummary.end(); ++c )
  {
    list.append( *c );
    list.append( coverageSummaries( c ) );
  }
  return list;
}
